#!/usr/bin/lua
--[[-------------------------------------------------------------------

  ChunkBake
  A Lua 5 binary chunk assembler.

  Copyright (c) 2005 Kein-Hong Man <khman@users.sf.net>
  The COPYRIGHT file describes the conditions under which this
  software may be distributed (basically a Lua 5-style license.)

  http://luaforge.net/projects/chunkbake/
  (TODO) http://www.geocities.com/keinhong/chunkbake.html
  See the ChangeLog for more information.

--]]-------------------------------------------------------------------

--[[-------------------------------------------------------------------
-- Notes:
-- * Modules aren't separate object; they mess with each other
-- * Numbers are parsed and converted into the internal number data type.
--   The accuracy of the output file's number data type will be
--   restricted by this internal number type.
-- * Handling of numbers is limited to the precision/range available.
-- * Floating point conversion may not be accurate to the last digit
--   because of the way the conversion function is written.
--   Untested for numbers that need to be specified very precisely.
-- * TODOs: please see the TODO files for a list of TODOs.
-- * For notes on assembler behaviour, please see the README file.
-----------------------------------------------------------------------
-- Modules: Number, Lex, Parse, Code
-- Outline: main() -> ChunkBakeDoFiles() -> Parse:Parse() -> Lex:Lex()
--]]-------------------------------------------------------------------

--[[-------------------------------------------------------------------
-- Description and help texts
--]]-------------------------------------------------------------------

title = [[
ChunkBake: A Lua 5 binary chunk assembler
Version 0.6.0 (20050305)  Copyright (c) 2005 Kein-Hong Man
The COPYRIGHT file describes the conditions under which this
software may be distributed (basically a Lua 5-style license.)
]]
-----------------------------------------------------------------------
USAGE = [[
usage: %s [options] [filenames]

options:
  -h, --help        prints usage information
  --quiet           do not display warnings
  -o <file>         specify file name to write output listing
  --                stop handling arguments

example:
  >%s myscript.asm -o myscript.out
]]
--TODO  --list            generate listing file
--TODO  --run             assemble and execute

-----------------------------------------------------------------------
-- Global constants, argument flags, etc.
-----------------------------------------------------------------------

config = {}
config.EXT_BIN = ".out"
config.EXT_LST = ".lst"

--[[-------------------------------------------------------------------
-- Other globals
--]]-------------------------------------------------------------------

arg_other = {}          -- other arguments (for --run option)

-- arg[0] is not set if ChunkBake.lua is embedded
local usage, exec
if arg[0] then exec = "lua ChunkBake.lua" else exec = "ChunkBake" end
usage = string.format(USAGE, exec, exec)

--[[-------------------------------------------------------------------
-- Number handler for multiple data types
-- * the actual call is set as Number:Convert(n)
-- * calls Code:Error() out of sheer laziness, beware!
--]]-------------------------------------------------------------------
Number = {}     -- number object

-----------------------------------------------------------------------
-- Selects a conversion function based on the given name or sample
-----------------------------------------------------------------------
function Number:SetNumberType(s)
  local matcher = {
    ["double"] = "double",
    ["single"] = "single",
    ["int"] = "int",
    ["long long"] = "long long",
    ["\182\9\147\104\231\245\125\65"] = "double",
    ["\59\175\239\75"] = "single",
    ["\118\94\223\1"] = "int",
    ["\118\94\223\1\0\0\0\0"] = "long long",
  }
  local test_number = {
    ["double"] = "\182\9\147\104\231\245\125\65",
    ["single"] = "\59\175\239\75",
    ["int"] = "\118\94\223\1",
    ["long long"] = "\118\94\223\1\0\0\0\0",
  }
  local match = matcher[s]
  if not match then
    Code:Error("unrecognized number type")
  end
  self.Convert = self[match]    -- initialize parameters
  self.test_number = test_number[match]
  self.lua_Number = string.len(self.test_number)
  return match
end

-----------------------------------------------------------------------
-- Support function for convert_to functions
-----------------------------------------------------------------------
function Number:grab_byte(v)
  return math.floor(v / 256), string.char(math.mod(math.floor(v), 256))
end

-----------------------------------------------------------------------
-- Converts a IEEE754 double number to an 8-byte little-endian string
-- * NOTE: see warning about accuracy in the header comments!
-----------------------------------------------------------------------
Number["double"] = function(self, x)
  local sign = 0
  if x < 0 then sign = 1; x = -x end
  local mantissa, exponent = math.frexp(x)
  if x == 0 then -- zero
    mantissa, exponent = 0, 0
  else
    mantissa = (mantissa * 2 - 1) * math.ldexp(0.5, 53)
    exponent = exponent + 1022
  end
  local v, byte = "" -- convert to bytes
  x = mantissa
  for i = 1,6 do
    x, byte = self:grab_byte(x); v = v..byte -- 47:0
  end
  x, byte = self:grab_byte(exponent * 16 + x); v = v..byte -- 55:48
  x, byte = self:grab_byte(sign * 128 + x); v = v..byte -- 63:56
  return v
end

-----------------------------------------------------------------------
-- Converts a IEEE754 single number to a 4-byte little-endian string
-- * TODO UNTESTED!!! *
-----------------------------------------------------------------------
Number["single"] = function(self, x)
  local sign = 0
  if x < 0 then sign = 1; x = -x end
  local mantissa, exponent = math.frexp(x)
  if x == 0 then -- zero
    mantissa = 0; exponent = 0
  else
    mantissa = (mantissa * 2 - 1) * math.ldexp(0.5, 24)
    exponent = exponent + 126
  end
  local v, byte = "" -- convert to bytes
  x, byte = self:grab_byte(mantissa); v = v..byte -- 7:0
  x, byte = self:grab_byte(x); v = v..byte -- 15:8
  x, byte = self:grab_byte(exponent * 128 + x); v = v..byte -- 23:16
  x, byte = self:grab_byte(sign * 128 + x); v = v..byte -- 31:24
  return v
end

-----------------------------------------------------------------------
-- Converts a number to a little-endian integer string
-- * TODO UNTESTED!!! *
-----------------------------------------------------------------------
Number["int"] = function(self, x)
  local v = ""
  x = math.floor(x)
  if x >= 0 then
    for i = 1, self.lua_Number do
      v = v..string.char(math.mod(x, 256)); x = math.floor(x / 256)
    end
  else-- x < 0
    x = -x
    local carry = 1
    for i = 1, self.lua_Number do
      local c = 255 - math.mod(x, 256) + carry
      if c == 256 then c = 0; carry = 1 else carry = 0 end
      v = v..string.char(c); x = math.floor(x / 256)
    end
  end
  -- optional overflow test
  if x > 0 then Code:Warn("integer number conversion overflow") end
  return v
end

-----------------------------------------------------------------------
-- * WARNING this will fail for large long longs (64-bit numbers)
--   because long longs exceeds the precision of doubles.
-----------------------------------------------------------------------
Number["long long"] = Number["int"]

--[[-------------------------------------------------------------------
-- (ChunkBake) Lua assembly language lexer
-- * currently does not lex double-character operators
-- * tries to be a proper lexer so that expression evaluation can
--   be added easily at a later date
-- * the handling of # immediate constants style is a bit messy
--]]-------------------------------------------------------------------
Lex = {}        -- lexer object
--OPCODE        -- this is set up by Lex:Init()

-----------------------------------------------------------------------
-- Initialize lexer
-- * sets up OPCODE as a global table
-----------------------------------------------------------------------
function Lex:Init(asmdata, asmfile)
  self.pos = 1
  self.line = 1
  self.asm = asmdata or ""
  self.asmlen = string.len(self.asm)
  self.asmfile = asmfile or "(no name)"
  if not self.EOF then self.EOF = -1 end
  self:nextc()
  ---------------------------------------------------------------
  -- build opcode name table
  ---------------------------------------------------------------
  local op =
    "MOVE LOADK LOADBOOL LOADNIL GETUPVAL \
    GETGLOBAL GETTABLE SETGLOBAL SETUPVAL SETTABLE \
    NEWTABLE SELF ADD SUB MUL \
    DIV POW UNM NOT CONCAT \
    JMP EQ LT LE TEST \
    CALL TAILCALL RETURN FORLOOP TFORLOOP \
    TFORPREP SETLIST SETLISTO CLOSE CLOSURE"
  if not self.OPCODE then
    self.OPCODE = {}
    local i = 0
    for v in string.gfind(op, "[^%s]+") do
      self.OPCODE[i] = v; self.OPCODE[v] = i; i = i + 1
    end
    self.OPCODE_COUNT = i
  end
  ---------------------------------------------------------------
  -- build keyword name table
  ---------------------------------------------------------------
  local kw = "nil false true"
  if not self.KEYWORD then
    self.KEYWORD = {}
    local i = 0
    for v in string.gfind(kw, "[^%s]+") do
      self.KEYWORD[i] = v; self.KEYWORD[v] = i; i = i + 1
    end
    self.KEYWORD_COUNT = i
  end
end

-----------------------------------------------------------------------
-- Format error message for display, throw an error
-----------------------------------------------------------------------
function Lex:Error(msg)
  error(string.format("%s:%d: Error: %s", self.asmfile, self.line, msg))
end

-----------------------------------------------------------------------
-- Get next character
-----------------------------------------------------------------------
function Lex:nextc()
  if self.pos > self.asmlen then
    self.c = self.EOF; self.ch = ""; self.dble = "" return
  end
  self.c = string.byte(self.asm, self.pos)      -- return a character
  self.pos = self.pos + 1
  -- universal handling for newlines, a newline is always one LF
  if self.c == 13 then
    if self.pos <= self.asmlen                  -- translate CRLF
       and string.byte(self.asm, self.pos) == 10 then
      self.pos = self.pos + 1
    end
    self.c = 10
  end
  self.ch = string.char(self.c)
  if self.pos <= self.asmlen then               -- easy dual-char
    self.c2 = string.byte(self.asm, self.pos)
    self.ch2 = string.char(self.c2)
    self.dble = string.char(self.c, self.c2)
  else
    self.dble = self.ch
  end
end

-----------------------------------------------------------------------
-- skip rest of a line (for single-line comments)
-----------------------------------------------------------------------
function Lex:SkipToEOL()
  while self.ch ~= "\n" and self.c ~= self.EOF do self:nextc() end
end

-----------------------------------------------------------------------
-- save character to buffer, get next character
-----------------------------------------------------------------------
function Lex:SaveNext()
  self.buff = self.buff..self.ch; self:nextc()
end

-----------------------------------------------------------------------
-- continue reading characters of one particular class
-----------------------------------------------------------------------
function Lex:ReadChars(pat)
  while string.find(self.ch, pat) do
    self.buff = self.buff..self.ch; self:nextc()
  end
end

-----------------------------------------------------------------------
-- read and convert a number
-----------------------------------------------------------------------
function Lex:ReadNumber()
  local isint = true
  self.buff = ""
  self:ReadChars("%d")                  -- integer part
  if self.dble == ".." then             -- prefer ..
    local num = tonumber(self.buff)
    return num, isint
  elseif self.ch == "." then            -- optional .
    isint = false; self:SaveNext()
  end
  self:ReadChars("%d")                  -- fraction part
  if string.upper(self.ch) == "E" then
    isint = false; self:SaveNext()
    if self.ch == "+" or self.ch == "-" then self:SaveNext() end
    self:ReadChars("%d")                -- exponent
  end
  local num = tonumber(self.buff)
  if not num then self:Error("could not lex number") end
  return num, isint
end

-----------------------------------------------------------------------
-- read and convert a hexadecimal number
-----------------------------------------------------------------------
function Lex:ReadHex()
  self.buff = ""
  self:nextc(); self:nextc()
  self:ReadChars("%x")                  -- read in hex digits
  local num = tonumber(self.buff, 16)
  if not num then self:Error("could not lex hex number") end
  return num, true
end

-----------------------------------------------------------------------
-- reads a single-quoted or double-quoted string
-----------------------------------------------------------------------
function Lex:ReadString(delimiter)
  self:nextc()
  self.buff = ""
  while self.ch ~= delimiter do
    -------------------------------------------------------------
    if self.c == self.EOF or self.ch == "\n" then
      self:Error("undelimited string")
    -------------------------------------------------------------
    elseif self.ch == "\\" and string.len(self.dble) == 2 then
      self:nextc()
      local i = string.find("\nabfnrtv", self.ch, 1, 1)
      if i then                                 -- standard escapes
        self.ch = string.sub("\n\a\b\f\n\r\t\v", i, i)
        if i == 1 then self.line = self.line + 1 end
        self:SaveNext()
      elseif string.find(self.ch, "%d") then    -- \xxx sequence
        local c, j = 0, 0
        repeat
          c = 10 * c + self.ch; self:nextc(); j = j + 1
        until (j >= 3 or not string.find(self.ch, "%d"))
        if c > 255 then
          self:Error("\\ddd escape overflow")
        end
        self.buff = self.buff..string.char(c)
      else                                      -- punctuation
        self:SaveNext()
      end
    -------------------------------------------------------------
    else                                      -- normal literal
      self:SaveNext()
    end--if self.ch
    -------------------------------------------------------------
  end--while
  self:nextc()
  return self.buff
end

-----------------------------------------------------------------------
-- reads a [[...]] style long string or comment
-----------------------------------------------------------------------
function Lex:ReadLongString(comment)
  local level = 0
  self.buff = ""
  self:nextc(); self:nextc()
  if self.ch == "\n" then     -- skip first newline
    self:nextc(); self.line = self.line + 1
  end
  while true do
    -------------------------------------------------------------
    if self.c == self.EOF then                  -- unexpected EOF
      if comment then
        self:Error("incomplete long comment")
      else
        self:Error("incomplete long string")
      end
    -------------------------------------------------------------
    elseif self.ch == "\n" then                 -- end-of-line
      self:SaveNext(); self.line = self.line + 1
    -------------------------------------------------------------
    elseif self.dble == "[[" then               -- inc nesting
      self:SaveNext(); self:SaveNext()
      level = level + 1
    -------------------------------------------------------------
    elseif self.dble == "]]" then               -- dec nesting
      self:nextc(); self:nextc()
      if level == 0 then break end              -- end of string
      self.buff = self.buff.."]]"
      level = level - 1
    -------------------------------------------------------------
    else                                        -- string content
      self:SaveNext()
    end--if self.ch
    -------------------------------------------------------------
  end--while
  return self.buff
end

-----------------------------------------------------------------------
-- Main lexer function
-----------------------------------------------------------------------
function Lex:Lex()
  ---------------------------------------------------------------
  -- main lexer loop
  ---------------------------------------------------------------
  while true do
    -------------------------------------------------------------
    if self.ch == "\n" then                     -- end-of-line
      self:nextc(); self.line = self.line + 1
      return "TK_EOL", "\n"
    -------------------------------------------------------------
    elseif self.ch == ";" then                  -- comment
      self:SkipToEOL()
    -------------------------------------------------------------
    elseif self.dble == "--" then               -- comment
      self:nextc(); self:nextc()
      if self.dble == "[[" then
        self:ReadLongString(true)
      else
        self:SkipToEOL()                        -- short comment
      end
    -------------------------------------------------------------
    elseif self.ch == "\"" or self.ch == "\'" then -- string
      return "TK_STRING", self:ReadString(self.ch)
    -------------------------------------------------------------
    elseif self.dble == "[[" then               -- long string
      return "TK_STRING", self:ReadLongString()
    -------------------------------------------------------------
    elseif self.dble == ".." then               -- concat operator
      self:nextc(); self:nextc()
      return "TK_CONCAT", ".."
    -------------------------------------------------------------
    elseif self.ch == "\\" then
      self:nextc()
      if self.ch == "\n" then                   -- line continue
        self:nextc(); self.line = self.line + 1
      else
        return "TK_OP", "\\"
      end
    -------------------------------------------------------------
    elseif self.c == self.EOF then              -- end-of-file
      return "TK_EOF", self.c
    -------------------------------------------------------------
    elseif string.find(self.ch, "%s") then      -- whitespace
      self:nextc()
    -------------------------------------------------------------
    elseif string.find(self.ch, "%d") then      -- number
      if self.ch2 == "x" or self.ch2 == "X" then
        return "TK_NUM", self:ReadHex()
      end
      return "TK_NUM", self:ReadNumber()
    -------------------------------------------------------------
    elseif string.find(self.ch, "[%a_]") then   -- symbols, etc.
      self.buff = ""; self:ReadChars("[%w_]")
      if self.ch == ":" then                    -- label:
        self:nextc()
        return "TK_LABEL", self.buff
      elseif string.find(self.buff, "^[rR][%d]+$") then   -- Rnum
        local reg = tonumber(string.sub(self.buff, 2))
        return "TK_REG", reg
      elseif self.OPCODE[string.upper(self.buff)] then  -- mnemonic
        return "TK_MNE", string.upper(self.buff)
      elseif self.KEYWORD[self.buff] then       -- keyword
        return "TK_KEY", self.buff
      end
      return "TK_SYM", self.buff                -- symbol
    -------------------------------------------------------------
    elseif self.ch == "$" then
      if string.find(self.ch2, "%d") then       -- $num
        self:nextc()
        self.buff = ""; self:ReadChars("%d")
        local reg = tonumber(self.buff)
        return "TK_REG", reg
      else
        self:nextc()
        return "TK_OP", "$"
      end
    -------------------------------------------------------------
    elseif self.ch == "." then
      if string.find(self.ch2, "%d") then       -- .number
        return "TK_NUM", self:ReadNumber()
      end
      self:nextc()
      self.buff = "."; self:ReadChars("[%w_]")  -- .directive
      return "TK_CMD", string.lower(self.buff)
    -------------------------------------------------------------
    elseif self.ch == "#" then                  -- #immediate
      self:nextc()
      if string.find(self.ch, "%d") then
        if self.ch2 == "x" or self.ch2 == "X" then
          return "TK_IMM", self:ReadHex()
        else
          return "TK_IMM", self:ReadNumber()
        end
      elseif self.ch == "-" and string.find(self.ch2, "%d") then
        -- TODO make # a separate prefix token for expressions
        self:nextc()
        local num, isint = self:ReadNumber()
        return "TK_IMM", -num, isint
      else
        return "TK_OP", "#"
      end
    -------------------------------------------------------------
    elseif string.find(self.ch, "%p") then      -- operators, etc.
      local punc = self.ch
      self:nextc()
      return "TK_OP", punc
    -------------------------------------------------------------
    else                                        -- catch the rest
      self:Error("unknown character or control character")
    end--if self.ch
    -------------------------------------------------------------
  end--while
end

--[[-------------------------------------------------------------------
-- (ChunkBake) Lua assembly language parser
-- * not very clean integration with Lex and Code objects
-- * many variables and functions are freely shared between objects
--]]-------------------------------------------------------------------
Parse = {}      -- parser object

-----------------------------------------------------------------------
-- get a token from the lexer
-----------------------------------------------------------------------
function Parse:GetToken()
  self.tt, self.tok, self.isint = Lex:Lex()
  self.line = Lex.line
  -- Parse.line points to the correct line for error messages
  if self.tt == "TK_EOL" then self.line = self.line - 1 end
end

-----------------------------------------------------------------------
-- Format error message for display, throw an error
-----------------------------------------------------------------------
function Parse:Error(msg)
  error(string.format("%s:%d: Error: %s", self.asmfile, self.line, msg))
end

-----------------------------------------------------------------------
-- Expects an end-of-line at the end of a directive or function
-----------------------------------------------------------------------
function Parse:EndOfLine()
  if self.tt == "TK_EOL" or self.tt == "TK_EOF" then
    self:GetToken(); return
  end
  self:Error("end of directive or instruction line expected")
end

-----------------------------------------------------------------------
-- Check if next token is EOL or EOF
-----------------------------------------------------------------------
function Parse:IsEOL()
  return self.tt == "TK_EOL" or self.tt == "TK_EOF"
end

-----------------------------------------------------------------------
-- Optionally consumes an equal sign
-----------------------------------------------------------------------
function Parse:OptionalEqual()
  if self.tt == "TK_OP" and self.tok == "=" then
    self:GetToken(); return true
  end
end

-----------------------------------------------------------------------
-- Optionally consumes a comma
-----------------------------------------------------------------------
function Parse:OptionalComma()
  if self.tt == "TK_OP" and self.tok == "," then
    self:GetToken(); return true
  end
end

-----------------------------------------------------------------------
-- Optionally consumes a "[", and compulsory "]" if it does
-----------------------------------------------------------------------
function Parse:OptionalBracketOpen()
  if self.tt == "TK_OP" and self.tok == "[" then
    self:GetToken(); return true
  end
end

function Parse:BracketClose()
  if self.tt == "TK_OP" and self.tok == "]" then
    self:GetToken(); return true
  end
  self:Error("closing square bracket expected")
end

-----------------------------------------------------------------------
-- Optionally consumes a comma or a concat (..)
-----------------------------------------------------------------------
function Parse:OptionalCommaConcat()
  if (self.tt == "TK_OP" and self.tok == ",")
     or self.tt == "TK_CONCAT" then
    self:GetToken(); return true
  end
end

-----------------------------------------------------------------------
-- Parse and decodes an R(x) operand
-- * allows TK_REG (Rnum, $num), TK_NUM, TK_SYM (locals or labels)
-- * R(x) and RK(x) tracks register usage for maxstacksize
-----------------------------------------------------------------------
function Parse:OperandR()
  if self.tt == "TK_REG" then
  elseif self.tt == "TK_NUM" then
    if not self.isint then
      self:Error("integer R(x) register operand expected")
    end
  elseif self.tt == "TK_SYM" then
    local label = Code:IsLabel(self.tok)
    if label then
      self.tok = Code:GetLabel(self.tok, "local")
    else
      self.tok = Code:GetLocal(self.tok)
    end
  else
    self:Error("R(x) operand expected")
  end
  local r = self.tok
  Code:CheckR(r)
  self:GetToken()
  if r >= Code.func.maxstacksize then Code.func.maxstacksize = r + 1 end
  return r
end

-----------------------------------------------------------------------
-- Parse and decodes an RK(x) operand
-- * allows the three types of tokens allowed by Parse:OperandR(), and
--   TK_IMM (immediate numerical const), TK_STRING (string const)
-----------------------------------------------------------------------
function Parse:OperandRK()
  if self.tt == "TK_REG" then
  elseif self.tt == "TK_NUM" then
    if not self.isint then
      self:Error("integer RK(x) register operand expected")
    end
    Code:CheckRK(self.tok)
    if self.tok >= Code.config.MAXSTACK then
      local actual = self.tok - Code.config.MAXSTACK
      if not Code:IsConst(actual) then
        self:Error("constant number "..actual.." not yet defined")
      end
    end
  elseif self.tt == "TK_IMM" or self.tt == "TK_STRING" then
    self.tok = Code:GetConst(self.tok) + Code.config.MAXSTACK
  elseif self.tt == "TK_SYM" then
    local const = Code:GetConstSym(self.tok)
    if const then
      self.tok = const + Code.config.MAXSTACK
    else
      local label = Code:IsLabel(self.tok)
      if label then
        if label == "const" then
          self.tok = Code:GetLabel(self.tok, "const") + Code.config.MAXSTACK
        else
          self.tok = Code:GetLabel(self.tok, "local")
        end
      else
        self.tok = Code:GetLocal(self.tok)
      end--if label
    end--if const
  elseif self.tt == "TK_KEY" then
    if self.tok == "true" or self.tok == "false" then
      self:Error(self.tok.." cannot be used as an RK(x) operand")
    end
    -- TK_KEY, nil
    self.tok = Code:GetConst(nil) + Code.config.MAXSTACK
  else
    self:Error("RK(x) operand expected")
  end--if self.tt
  local r = self.tok
  Code:CheckRK(r)
  self:GetToken()
  if r < Code.config.MAXSTACK and
     r >= Code.func.maxstacksize then Code.func.maxstacksize = r + 1 end
  return r
end

-----------------------------------------------------------------------
-- Parse and decodes a Kst(x) operand
-- * allows TK_SYM (labeled const), TK_NUM (constant number),
--   TK_IMM (immediate number), TK_STRING (immediate string)
-----------------------------------------------------------------------
function Parse:OperandKst()
  if self.tt == "TK_NUM" then
    if not self.isint then
      self:Error("integer Kst(x) constant operand expected")
    end
    if not Code:IsConst(self.tok) then
      self:Error("constant number "..self.tok.." not yet defined")
    end
  elseif self.tt == "TK_IMM" or self.tt == "TK_STRING" then
    self.tok = Code:GetConst(self.tok)
  elseif self.tt == "TK_SYM" then
    local const = Code:GetConstSym(self.tok)
    if const then
      self.tok = const
    else
      self.tok = Code:GetLabel(self.tok, "const")
    end
  elseif self.tt == "TK_KEY" then
    if self.tok == "true" or self.tok == "false" then
      self:Error(self.tok.." cannot be used as a Kst(x) operand")
    end
    -- TK_KEY, nil
    self.tok = Code:GetConst(nil)
  else
    self:Error("Kst(x) operand expected")
  end--if self.tt
  local k = self.tok
  Code:CheckKst(k)
  self:GetToken()
  return k
end

-----------------------------------------------------------------------
-- Parse and decodes an integer number, optionally checks range
-- * if range is a string, then allows immediates also, this is only for
--   the NEWTABLE instruction, for real array and hash sizes
-----------------------------------------------------------------------
function Parse:OperandNum(range)
  if self.tt == "TK_NUM" or
     (type(range) == "string" and self.tt == "TK_IMM") then
    if not self.isint then self:Error("integer operand expected") end
  else
    self:Error("integer operand expected")
  end
  local val, typ = self.tok, self.tt
  self:GetToken()
  if type(range) == "number" and (val < 0 or val >= range) then
    self:Error("numeric operand out of range")
  end
  return val, typ
end

-----------------------------------------------------------------------
-- Parse and decodes a displacement or label
-- * can be TK_NUM (absolute), TK_IMM (relative), or TK_SYM
-- * if label is a forward reference, it is kept as a string and is
--   resolved at the end of a function's parsing (.end)
-----------------------------------------------------------------------
function Parse:OperandDisp()
  local disp = self.tok
  if self.tt == "TK_NUM" then
    if not self.isint then
      self:Error("integer displacement absolute operand expected")
    end
  elseif self.tt == "TK_IMM" then
    if not self.isint then
      self:Error("integer displacement relative operand expected")
    end
  elseif self.tt == "TK_SYM" then
    disp = Code:GetLabel(disp, "code")
  else
    self:Error("displacement or label expected")
  end
  if type(disp) == "number" and self.tt ~= "TK_IMM" then
    -- at time of parsing, PC is positioned at the last instruction
    -- so there is a difference of 2 compared to the next instruction
    disp = disp - Code.func.sizecode - 2
  end
  Code:CheckDisp(disp)
  self:GetToken()
  return disp
end

-----------------------------------------------------------------------
-- Parse a key-value parameter pair; SYNTAX: <key>=<value>
-----------------------------------------------------------------------
function Parse:Parameter()
  if self.tt ~= "TK_SYM" then
    self:Error("symbol expected for parameter key")
  end
  local pkey = self.tok
  self:GetToken()
  if not(self.tt == "TK_OP" and self.tok == "=") then
    self:Error("'=' expected in key-value parameter pair")
  end
  self:GetToken()
  if not(self.tt == "TK_NUM" or self.tt == "TK_STRING") then
    self:Error("number or string expected as parameter value")
  end
  local pval = self.tok
  self:GetToken()
  return pkey, pval
end

-----------------------------------------------------------------------
-- Parse a directive or command
-----------------------------------------------------------------------
function Parse:Directive(cmd)
  ---------------------------------------------------------------
  -- declares global header properties
  ---------------------------------------------------------------
  if cmd == ".header" then
    if Code:HasLabels() then
      self:Error(cmd.." doesn't accept label prefixes")
    end
    while not self:IsEOL() do
      local pkey, pval = self:Parameter()
      Code:HeaderOption(pkey, pval)
      self:OptionalComma()
    end
    self:EndOfLine()
  ---------------------------------------------------------------
  -- declares a function prototype
  ---------------------------------------------------------------
  elseif cmd == ".function" or cmd == ".func" then
    Code:InitHeader()
    Code:FunctionBegin()
    while not self:IsEOL() do
      local pkey, pval = self:Parameter()
      Code:FunctionOption(pkey, pval)
      self:OptionalComma()
    end
    self:EndOfLine()
    if Code:HasLabels() then
      if Code.func.prev then
        Code:SetLabel("function", Code.func.prev.sizep, 1)
      else
        self:Error("top-level function cannot have a label")
      end
    end
  ---------------------------------------------------------------
  -- ends a function
  ---------------------------------------------------------------
  elseif cmd == ".end" then
    if Code:HasLabels() then
      self:Error(cmd.." doesn't accept label prefixes")
    end
    -- TODO possible: ".end TK_SYM" function syntax
    --if self.tt == "TK_SYM" then <blah blah> end
    self:EndOfLine()
    Code:FunctionEnd()
  ---------------------------------------------------------------
  -- declares a parameter (more or less a local)
  ---------------------------------------------------------------
  elseif cmd == ".param" then
    if Code.state ~= 1 then
      self:Error(".param declaration illegal outside a function")
    end
    if self.tt == "TK_SYM" then
      local param, loc = self.tok
      self:GetToken()
      -- SYNTAX: .param <symbol> TK_EOL
      loc = Code:AddParam(param)
      self:EndOfLine()
      if Code:HasLabels() then Code:SetLabel("local", loc) end
    else
      self:Error("symbol expected in .param directive")
    end
  ---------------------------------------------------------------
  -- declares a local variable
  -- * TODO doesn't handle startpc/endpc yet
  ---------------------------------------------------------------
  elseif cmd == ".local" then
    if Code.state ~= 1 then
      self:Error(".local declaration illegal outside a function")
    end
    if self.tt == "TK_SYM" then
      local locname, loc = self.tok
      self:GetToken()
      if self:OptionalComma() then
        if self.tt == "TK_NUM" and self.isint then
          -- SYNTAX: .local <symbol>, <number> TK_EOL
          loc = Code:AddLocal(locname, self.tok)
          self:GetToken()
        else
          self:Error("number expected after comma in .local directive")
        end
      else-- no comma
        if self.tt == "TK_NUM" and self.isint then
          -- SYNTAX: .local <symbol> <number> TK_EOL
          loc = Code:AddLocal(locname, self.tok)
          self:GetToken()
        elseif self:IsEOL() then
          -- SYNTAX: .local <symbol> TK_EOL
          loc = Code:AddLocal(locname)
        else
          self:Error("number expected as second operand in .local directive")
        end
      end
      self:EndOfLine()
      if Code:HasLabels() then Code:SetLabel("local", loc) end
    else
      self:Error("symbol expected in .local directive")
    end
  ---------------------------------------------------------------
  -- declares an upvalue
  ---------------------------------------------------------------
  elseif cmd == ".upvalue" then
    if Code.state ~= 1 then
      self:Error(".upvalue declaration illegal outside a function")
    end
    if self.tt == "TK_SYM" then
      local upval, upv = self.tok
      self:GetToken()
      if self:OptionalComma() then
        if self.tt == "TK_NUM" and self.isint then
          -- SYNTAX: .upvalue <symbol>, <number> TK_EOL
          upv = Code:AddUpvalue(upval, self.tok)
          self:GetToken()
        else
          self:Error("number expected after comma in .upvalue directive")
        end
      else-- no comma
        if self.tt == "TK_NUM" and self.isint then
          -- SYNTAX: .upvalue <symbol> <number> TK_EOL
          upv = Code:AddUpvalue(upval, self.tok)
          self:GetToken()
        elseif self:IsEOL() then
          -- SYNTAX: .upvalue <symbol> TK_EOL
          upv = Code:AddUpvalue(upval)
        else
          self:Error("number expected as second operand in .upvalue directive")
        end
      end
      self:EndOfLine()
      if Code:HasLabels() then Code:SetLabel("upvalue", upv) end
    else
      self:Error("symbol expected in .upvalue directive")
    end
  ---------------------------------------------------------------
  -- declares a constant, with optional label
  ---------------------------------------------------------------
  elseif cmd == ".const" then
    if Code.state ~= 1 then
      self:Error(".const declaration illegal outside a function")
    end
    -- SYNTAX: .const [<symbol>,] ...
    local const
    if self.tt == "TK_SYM" then
      const = self.tok
      self:GetToken()
    end
    self:OptionalComma()
    if self.tt == "TK_NUM" or self.tt == "TK_IMM" or
       self.tt == "TK_STRING" or self.tt == "TK_KEY" then
      local val, cn = self.tok
      -- handle nil constants
      if self.tt == "TK_KEY" then
        if val == "true" or val == "false" then
          self:Error(val.." cannot be used in .const directive")
        end
        val = nil
      end
      self:GetToken()
      if self:OptionalComma() then
        if self.tt == "TK_NUM" and self.isint then
          -- SYNTAX: .const [<symbol>,] <num>|<imme>|<string>|nil, <number> TK_EOL
          cn = Code:AddConst(const, val, self.tok)
          self:GetToken()
        else
          self:Error("number expected after comma in .const directive")
        end
      else-- no comma
        if self.tt == "TK_NUM" and self.isint then
          -- SYNTAX: .const [<symbol>,] <num>|<imme>|<string>|nil <number> TK_EOL
          cn = Code:AddConst(const, val, self.tok)
          self:GetToken()
        elseif self:IsEOL() then
          -- SYNTAX: .const [<symbol>,] <num>|<imme>|<string> TK_EOL
          cn = Code:AddConst(const, val)
        else
          self:Error("number operand expected in .const directive")
        end
      end
      self:EndOfLine()
      if Code:HasLabels() then Code:SetLabel("const", cn) end
    else
      self:Error("number or string expected in .const directive")
    end
  ---------------------------------------------------------------
  else
    self:Error("unknown directive "..cmd.." encountered")
  end--if cmd
end

-----------------------------------------------------------------------
-- Parse an instruction
-----------------------------------------------------------------------
function Parse:Instruction(mne)
  local RA, RB, RC, RBx
  if Code.state ~= 1 then
    self:Error("instruction cannot live outside a function")
  end
  ---------------------------------------------------------------
  -- MOVE R(A) [,] R(B)
  ---------------------------------------------------------------
  if mne == "MOVE" then
    RA = self:OperandR()
    self:OptionalComma()
    RB = self:OperandR()
    self:EndOfLine()
    Code:AddInst(Code:iABC(mne, RA, RB, 0))
  ---------------------------------------------------------------
  -- LOADK R(A) [,] Kst(Bx)
  ---------------------------------------------------------------
  elseif mne == "LOADK" then
    RA = self:OperandR()
    self:OptionalComma()
    RBx = self:OperandKst()
    self:EndOfLine()
    Code:AddInst(Code:iABx(mne, RA, RBx))
  ---------------------------------------------------------------
  -- LOADBOOL R(A) [,] B(0|1|true|false) [,] C(0|1)
  ---------------------------------------------------------------
  elseif mne == "LOADBOOL" then
    RA = self:OperandR()
    self:OptionalComma()
    if self.tt == "TK_NUM" and self.isint and
       (self.tok == 0 or self.tok == 1) then
      RB = self.tok
    elseif self.tt == "TK_KEY" and
       (self.tok == "true" or self.tok == "false") then
      if self.tok == "true" then RB = 1 else RB = 0 end
    elseif self:IsEOL() then
      self:Error("missing operand(s)")
    else
      self:Error("operand B must be 0, 1, true or false")
    end
    self:GetToken()
    self:OptionalComma()
    if self.tt == "TK_NUM"  and self.isint and
       (self.tok == 0 or self.tok == 1) then
      RC = self.tok
    elseif self:IsEOL() then
      self:Error("missing operand(s)")
    else
      self:Error("operand C must be 0 or 1")
    end
    self:GetToken()
    self:EndOfLine()
    Code:AddInst(Code:iABC(mne, RA, RB, RC))
  ---------------------------------------------------------------
  -- LOADNIL R(A) [(,|..)] R(B)
  ---------------------------------------------------------------
  elseif mne == "LOADNIL" then
    RA = self:OperandR()
    self:OptionalCommaConcat()
    RB = self:OperandR()
    self:EndOfLine()
    if RA > RB then
      self:Error("operand A must not exceed operand B")
    end
    Code:AddInst(Code:iABC(mne, RA, RB, 0))
  ---------------------------------------------------------------
  -- GETUPVAL R(A) [,] Upvalue[Bx]
  -- SETUPVAL R(A) [,] Upvalue[Bx]
  ---------------------------------------------------------------
  elseif mne == "GETUPVAL"
      or mne == "SETUPVAL" then
    RA = self:OperandR()
    self:OptionalComma()
    if self.tt == "TK_NUM" and self.isint then
      RBx = self.tok
      if not Code:IsUpvalue(RBx) then
        self:Error("undefined upvalue number "..RBx)
      end
    elseif self.tt == "TK_SYM" then
      if Code:IsLabel(self.tok) then
        RBx = Code:GetLabel(self.tok, "upvalue")
      else
        RBx = Code:GetUpvalue(self.tok)
      end
    elseif self:IsEOL() then
      self:Error("missing operand(s)")
    else
      self:Error("integer number or upvalue operand expected")
    end
    self:GetToken()
    self:EndOfLine()
    Code:AddInst(Code:iABx(mne, RA, RBx))
  ---------------------------------------------------------------
  -- GETGLOBAL R(A) [,] Kst(Bx)
  -- SETGLOBAL R(A) [,] Kst(Bx)
  ---------------------------------------------------------------
  elseif mne == "GETGLOBAL"
      or mne == "SETGLOBAL" then
    RA = self:OperandR()
    self:OptionalComma()
    RBx = self:OperandKst()
    self:EndOfLine()
    Code:AddInst(Code:iABx(mne, RA, RBx))
  ---------------------------------------------------------------
  -- GETTABLE R(A) [,] R(B) [,] RK(C)
  -- GETTABLE R(A) [,] R(B) "[" RK(C) "]"
  ---------------------------------------------------------------
  elseif mne == "GETTABLE" then
    RA = self:OperandR()
    self:OptionalComma()
    RB = self:OperandR()
    if self:OptionalBracketOpen() then
      RC = self:OperandRK()
      self:BracketClose()
    else
      self:OptionalComma()
      RC = self:OperandRK()
    end
    self:EndOfLine()
    Code:AddInst(Code:iABC(mne, RA, RB, RC))
  ---------------------------------------------------------------
  -- SETTABLE R(A) [,] RK(B) [,] RK(C)
  -- SETTABLE R(A) "[" RK(B) "]" [,] RK(C)
  ---------------------------------------------------------------
  elseif mne == "SETTABLE" then
    RA = self:OperandR()
    if self:OptionalBracketOpen() then
      RB = self:OperandRK()
      self:BracketClose()
    else
      self:OptionalComma()
      RB = self:OperandRK()
    end
    self:OptionalComma()
    RC = self:OperandRK()
    self:EndOfLine()
    Code:AddInst(Code:iABC(mne, RA, RB, RC))
  ---------------------------------------------------------------
  -- NEWTABLE R(A) [,] (num|imm) [,] (num|imm)
  ---------------------------------------------------------------
  elseif mne == "NEWTABLE" then
    local RBf, RCf
    RA = self:OperandR()
    self:OptionalComma()
    RB, RBf = self:OperandNum("")
    self:OptionalComma()
    RC, RCf = self:OperandNum("")
    self:EndOfLine()
    if RBf == "TK_IMM" then
      -- calculate actual array part size
      -- * conversion into a "floating point byte", where:
      --   value = (xxx) * 2^(mmmmm), byte (mmmmmxxx) in binary
      local m = 0
      while RB >= 8 do
        RB = math.floor((RB + 1) / 2); m = m + 1
      end
      RB = m * 8 + RB
    end
    if RCf == "TK_IMM" then
      -- calculate actual hash part size
      -- * the exp part of math.frexp() gives the exact amount
      RCf, RC = math.frexp(RC)
    end
    Code:AddInst(Code:iABC(mne, RA, RB, RC))
  ---------------------------------------------------------------
  -- SELF R(A) [,] R(B) [,] RK(C)
  -- SELF R(A) [,] R(B) "[" RK(C) "]"
  ---------------------------------------------------------------
  elseif mne == "SELF" then
    RA = self:OperandR()
    self:OptionalComma()
    RB = self:OperandR()
    if self:OptionalBracketOpen() then
      RC = self:OperandRK()
      self:BracketClose()
    else
      self:OptionalComma()
      RC = self:OperandRK()
    end
    self:EndOfLine()
    Code:AddInst(Code:iABC(mne, RA, RB, RC))
  ---------------------------------------------------------------
  -- ADD R(A) [,] RK(B) [,] RK(C)
  ---------------------------------------------------------------
  elseif mne == "ADD"
      or mne == "SUB"
      or mne == "MUL"
      or mne == "DIV"
      or mne == "POW" then
    RA = self:OperandR()
    self:OptionalComma()
    RB = self:OperandRK()
    self:OptionalComma()
    RC = self:OperandRK()
    self:EndOfLine()
    Code:AddInst(Code:iABC(mne, RA, RB, RC))
  ---------------------------------------------------------------
  -- UNM R(A) [,] R(B)
  ---------------------------------------------------------------
  elseif mne == "UNM"
      or mne == "NOT" then
    RA = self:OperandR()
    self:OptionalComma()
    RB = self:OperandR()
    self:EndOfLine()
    Code:AddInst(Code:iABC(mne, RA, RB, 0))
  ---------------------------------------------------------------
  -- CONCAT R(A) [,] R(B) [(,|..)] R(C)
  ---------------------------------------------------------------
  elseif mne == "CONCAT" then
    RA = self:OperandR()
    self:OptionalComma()
    RB = self:OperandR()
    self:OptionalCommaConcat()
    RC = self:OperandR()
    self:EndOfLine()
    if RB > RC then
      self:Error("operand B must not exceed operand C")
    end
    Code:AddInst(Code:iABC(mne, RA, RB, RC))
  ---------------------------------------------------------------
  -- EQ (0|1) [,] RK(B) [,] RK(C)
  ---------------------------------------------------------------
  elseif mne == "EQ"
      or mne == "LT"
      or mne == "LE" then
    if self.tt == "TK_NUM"  and self.isint and
       (self.tok == 0 or self.tok == 1) then
      RA = self.tok
    elseif self:IsEOL() then
      self:Error("missing operand(s)")
    else
      self:Error("operand A must be 0 or 1")
    end
    self:GetToken()
    self:OptionalComma()
    RB = self:OperandRK()
    self:OptionalComma()
    RC = self:OperandRK()
    self:EndOfLine()
    Code:AddInst(Code:iABC(mne, RA, RB, RC))
  ---------------------------------------------------------------
  -- TEST R(A) [,] R(B) [,] (0|1)
  ---------------------------------------------------------------
  elseif mne == "TEST" then
    RA = self:OperandR()
    self:OptionalComma()
    RB = self:OperandR()
    self:OptionalComma()
    if self.tt == "TK_NUM"  and self.isint and
       (self.tok == 0 or self.tok == 1) then
      RC = self.tok
    elseif self:IsEOL() then
      self:Error("missing operand(s)")
    else
      self:Error("operand C must be 0 or 1")
    end
    self:GetToken()
    self:EndOfLine()
    Code:AddInst(Code:iABC(mne, RA, RB, RC))
  ---------------------------------------------------------------
  -- JMP label|disp
  ---------------------------------------------------------------
  elseif mne == "JMP" then
    RBx = self:OperandDisp()
    self:EndOfLine()
    Code:AddInst(Code:iAsBx(mne, 0, RBx))
  ---------------------------------------------------------------
  -- CALL R(A) [,] B [,] C
  ---------------------------------------------------------------
  elseif mne == "CALL" then
    RA = self:OperandR()
    self:OptionalComma()
    RB = self:OperandNum(Code.MASK_B)
    self:OptionalComma()
    RC = self:OperandNum(Code.MASK_C)
    self:EndOfLine()
    Code:AddInst(Code:iABC(mne, RA, RB, RC))
  ---------------------------------------------------------------
  -- TAILCALL R(A) [,] B
  ---------------------------------------------------------------
  elseif mne == "TAILCALL" then
    RA = self:OperandR()
    self:OptionalComma()
    RB = self:OperandNum(Code.MASK_B)
    self:EndOfLine()
    Code:AddInst(Code:iABC(mne, RA, RB, 0))
  ---------------------------------------------------------------
  -- RETURN R(A) [,] B
  ---------------------------------------------------------------
  elseif mne == "RETURN" then
    RA = self:OperandR()
    self:OptionalComma()
    RB = self:OperandNum(Code.MASK_B)
    self:EndOfLine()
    Code:AddInst(Code:iABC(mne, RA, RB, 0))
  ---------------------------------------------------------------
  -- FORLOOP R(A) [,] label|disp
  ---------------------------------------------------------------
  elseif mne == "FORLOOP" then
    RA = self:OperandR()
    self:OptionalComma()
    RBx = self:OperandDisp()
    self:EndOfLine()
    Code:AddInst(Code:iAsBx(mne, RA, RBx))
  ---------------------------------------------------------------
  -- TFORLOOP R(A) [,] C
  ---------------------------------------------------------------
  elseif mne == "TFORLOOP" then
    RA = self:OperandR()
    self:OptionalComma()
    RC = self:OperandNum(Code.MASK_C)
    self:EndOfLine()
    Code:AddInst(Code:iABC(mne, RA, 0, RC))
  ---------------------------------------------------------------
  -- TFORPREP R(A) [,] label|disp
  ---------------------------------------------------------------
  elseif mne == "TFORPREP" then
    RA = self:OperandR()
    self:OptionalComma()
    RBx = self:OperandDisp()
    self:EndOfLine()
    Code:AddInst(Code:iAsBx(mne, RA, RBx))
  ---------------------------------------------------------------
  -- SETLIST R(A) [,] Bx
  -- SETLIST R(A) [,] start [..] end
  ---------------------------------------------------------------
  elseif mne == "SETLIST"
      or mne == "SETLISTO" then
    RA = self:OperandR()
    self:OptionalComma()
    RBx = self:OperandNum(Code.MASK_Bx)
    if self.tt == "TK_CONCAT" then
      self:GetToken()
      RC = self:OperandNum(Code.MASK_Bx)
      -- range calculation for RBx -> RC
      if math.mod(RBx, Code.config.FPF) ~= 1 then -- start index
        self:Error("range must start with a multiple of FPF ("..
                   Code.config.FPF..") plus 1")
      end
      if (RC - RBx) >= Code.config.FPF then -- range
        self:Error("range must be not exceed FPF ("..Code.config.FPF..")")
      end
      RBx = RC - 1
    end
    self:EndOfLine()
    Code:AddInst(Code:iABx(mne, RA, RBx))
  ---------------------------------------------------------------
  -- CLOSE R(A)
  ---------------------------------------------------------------
  elseif mne == "CLOSE" then
    RA = self:OperandR()
    self:EndOfLine()
    Code:AddInst(Code:iABC(mne, RA, 0, 0))
  ---------------------------------------------------------------
  -- CLOSURE R(A) [,] Bx
  ---------------------------------------------------------------
  elseif mne == "CLOSURE" then
    RA = self:OperandR()
    self:OptionalComma()
    if self.tt == "TK_NUM" then
      RBx = self.tok
      if RBx >= Code.func.sizep then
        self:Error("function number "..RBx.." not yet defined")
      end
    elseif self.tt == "TK_SYM" then
      RBx = Code:GetFunction(self.tok)
    elseif self:IsEOL() then
      self:Error("missing operand(s)")
    else
      self:Error("number or label expected")
    end
    self:GetToken()
    self:EndOfLine()
    Code:AddInst(Code:iABx(mne, RA, RBx))
  ---------------------------------------------------------------
  else
    -- should never arrive here, since all in op table are valid
    self:Error("unknown mnemonic '"..mne.."' encountered")
  end--if mne
  -- assign labels if any
  if Code:HasLabels() then Code:SetLabel("code", Code.func.sizecode) end
end

-----------------------------------------------------------------------
-- Parse a statement
-----------------------------------------------------------------------
function Parse:Statement()
  local tok = self.tok
  if self.tt == "TK_CMD" then
    self:GetToken()
    self:Directive(tok)
  elseif self.tt == "TK_MNE" then
    Code.mneline = Lex.line
    self:GetToken()
    self:Instruction(tok)
  else
    self:Error("directive or mnemonic expected")
  end
end

-----------------------------------------------------------------------
-- Main parser function
-----------------------------------------------------------------------
function Parse:Parse(asm, asmfile)
  self.asm = asm
  self.asmfile = asmfile
  Lex:Init(asm, asmfile)
  Code:Init(asmfile)
  self:GetToken()

  ---------------------------------------------------------------
  -- main parser loop
  -- * each loop parses one syntactical line (strings etc. might
  --   cause a "single" line to spill over into many lines
  ---------------------------------------------------------------
  repeat
    if self.tt == "TK_EOL" then -- empty line, do nothing
      self:GetToken()
    elseif self.tt == "TK_EOF" then
      -- nothing to do
    elseif self.tt == "TK_LABEL" then
      Code:AddLabel(self.tok)
      self:GetToken()
      if self.tt ~= "TK_EOL" and self.tt ~= "TK_EOF" then
        self:Statement()
      end
    elseif self.tt == "TK_SYM" then
      local maybeLabel = self.tok
      self:GetToken()
      if self.tt == "TK_EOL" then
        self:Error("orphaned symbol or identifier '"..maybeLabel.."'")
      end
      Code:AddLabel(maybeLabel)
      self:Statement()
    else
      self:Statement()
    end
  until self.tt == "TK_EOF"
  return Code:Dump()
end

--[[-------------------------------------------------------------------
-- (ChunkBake) Lua assembly language code generator
-- * state interpreted as follows:
--   0=global header, 1=after first function, 2=top-level closed
--]]-------------------------------------------------------------------
Code = {}       -- code generator object

-----------------------------------------------------------------------
-- Initialize code generator object
-----------------------------------------------------------------------
function Code:Init(asmfile)
  self.asmfile = asmfile
  self.state = 0
  self.level = 0        -- function levels are always >= 1
  self.root = nil       -- top-level function
  self.func = nil       -- current function being assembled
  self.labelsnow = {}   -- labels to process (per-line)
  self.header = {}      -- properties set by .header directives
  self.funcop = {}      -- properties set by .function directives
  self.config = nil     -- reset configuration
end

-----------------------------------------------------------------------
-- Format error message for display, throw an error
-----------------------------------------------------------------------
function Code:Error(msg)
  error(string.format("%s:%d: Error: %s", self.asmfile, Parse.line, msg))
end

-----------------------------------------------------------------------
-- Warning message for display
-----------------------------------------------------------------------
function Code:Warn(msg)
  if config.QUIET then return end
  local message = string.format("ChunkBake: %s:%d: Warning: %s\n", self.asmfile, Parse.line, msg)
  if not TEST then
    io.stderr:write(message)
    io.stderr:flush()
  else
    WARN = (WARN or "")..message
  end
end

-----------------------------------------------------------------------
-- Process setting of header properties
-- * case insensitive, so constants here must be *lower case*
-----------------------------------------------------------------------
function Code:HeaderOption(pkey, pval)
  if self.state ~= 0 then
    self:Error(".header declarations must be defined at the start")
  end
  pkey = string.lower(pkey)
  ---------------------------------------------------------------
  -- convenient test functions for .header property declarations
  ---------------------------------------------------------------
  local function Expect(typ)
    if type(pval) ~= typ then
      self:Error(typ.." expected for property "..pkey)
    end
  end
  ---------------------------------------------------------------
  local function Range(lo, hi)
    local n = pval
    if math.floor(n) ~= n then
      self:Error("integer number expected for property "..pkey)
    elseif (lo and n < lo) or (hi and n > hi) then
      self:Error("integer out of range for property "..pkey)
    end
  end
  ---------------------------------------------------------------
  -- match, check and set properties
  ---------------------------------------------------------------
  if pkey == "signature" then
    Expect("string")
    if string.len(pval) ~= 4 then
      self:Warn("signature declared in not of normal length (4)")
    end
    self.header.signature = pval
  ---------------------------------------------------------------
  elseif pkey == "version" then
    Expect("number"); Range(80)
    self.header.version = pval
  ---------------------------------------------------------------
  elseif pkey == "endianness" then
    Expect("number"); Range(0, 1)
    self.header.endianness = pval
  ---------------------------------------------------------------
  elseif pkey == "int" then
    Expect("number"); Range(4, 8)
    self.header.int = pval
  ---------------------------------------------------------------
  elseif pkey == "size_t" then
    Expect("number"); Range(2, 8)
    self.header.size_t = pval
  ---------------------------------------------------------------
  elseif pkey == "instruction" then
    Expect("number"); Range(4, 8)
    self.header.Instruction = pval
  ---------------------------------------------------------------
  elseif pkey == "size_op" then
    Expect("number"); Range(6, 64)
    self.header.SIZE_OP = pval
  ---------------------------------------------------------------
  elseif pkey == "size_a" then
    Expect("number"); Range(1, 64)
    self.header.SIZE_A = pval
  ---------------------------------------------------------------
  elseif pkey == "size_b" then
    Expect("number"); Range(1, 64)
    self.header.SIZE_B = pval
  ---------------------------------------------------------------
  elseif pkey == "size_c" then
    Expect("number"); Range(1, 64)
    self.header.SIZE_C = pval
  ---------------------------------------------------------------
  elseif pkey == "number_type" then
    Expect("string")
    if pval == "double" or pval == "single" or
       pval == "int" or pval == "long long" then
      self.header.number_type = Number:SetNumberType(pval)
    else
      self:Error("unrecognized number type for property "..pkey)
    end
  ---------------------------------------------------------------
  elseif pkey == "maxstack" then
    Expect("number"); Range(1)
    self.header.MAXSTACK = pval
  ---------------------------------------------------------------
  elseif pkey == "maxvars" then
    Expect("number"); Range(1)
    self.header.MAXVARS = pval
  ---------------------------------------------------------------
  elseif pkey == "maxupvalues" then
    Expect("number"); Range(1)
    self.header.MAXUPVALUES = pval
  ---------------------------------------------------------------
  elseif pkey == "maxparams" then
    Expect("number"); Range(1)
    self.header.MAXPARAMS = pval
  ---------------------------------------------------------------
  elseif pkey == "fpf" then
    Expect("number"); Range(2)
    self.header.FPF = pval
  ---------------------------------------------------------------
  else
    self:Error("unrecognized .header property "..pkey)
  end--if pkey
end

-----------------------------------------------------------------------
-- Once-only initialization of binary chunk properties and opcode stuff
-----------------------------------------------------------------------
function Code:InitHeader()
  self.state = 1 -- no longer accept .header directives
  if self.config then return end
  self.config = {}
  ---------------------------------------------------------------
  -- set global header values, default or user override
  -- * platform defaults are used if no override declared
  ---------------------------------------------------------------
  local SAMPLE = string.dump(function() end)
  local function Byte(i) return string.byte(SAMPLE, i + 1) end
  -- grab defaults from a sample binary chunk
  self.config.signature = self.header.signature or string.sub(SAMPLE, 1, 4)
  self.config.version = self.header.version or Byte(4)
  self.config.endianness = self.header.endianness or Byte(5)
  -- byte sizes
  self.config.int = self.header.int or Byte(6)
  self.config.size_t = self.header.size_t or Byte(7)
  self.config.Instruction = self.header.Instruction or Byte(8)
  -- bit sizes
  self.config.SIZE_OP = self.header.SIZE_OP or Byte(9)
  self.config.SIZE_A = self.header.SIZE_A or Byte(10)
  self.config.SIZE_B = self.header.SIZE_B or Byte(11)
  self.config.SIZE_C = self.header.SIZE_C or Byte(12)
  -- check validity
  if self.config.SIZE_OP < 6 then
    self:Error("operand field too small to hold all opcodes")
  end
  if self.config.SIZE_A >= self.config.SIZE_B then
    self:Error("bit field A must be smaller than bit field B and C")
  end
  if self.config.SIZE_B ~= self.config.SIZE_C then
    self:Error("bit field B must be the same as bit field C")
  end
  if (self.config.Instruction * 8) ~= (self.config.SIZE_OP +
      self.config.SIZE_A + self.config.SIZE_B + self.config.SIZE_C) then
    self:Error("instruction bit field sizes inconsistent with instruction size")
  end
  ---------------------------------------------------------------
  -- initialize number conversion function and settings
  ---------------------------------------------------------------
  self.config.number_type = self.header.number_type
    or Number:SetNumberType(string.sub(SAMPLE, 15, Byte(13) + 14))
  self.config.test_number = Number.test_number
  self.config.lua_Number = Number.lua_Number
  if not self.config.number_type then
    self:Error("number type initialization failed")
  end
  ---------------------------------------------------------------
  -- initialize constants for opcode handling
  ---------------------------------------------------------------
  self.config.SIZE_Bx = self.config.SIZE_B + self.config.SIZE_C
  self.MASK_OP = math.ldexp(1, self.config.SIZE_OP)
  self.MASK_B  = math.ldexp(1, self.config.SIZE_B)
  self.MASK_C  = math.ldexp(1, self.config.SIZE_C)
  self.MASK_Bx = math.ldexp(1, self.config.SIZE_Bx)
  self.MASK_A  = math.ldexp(1, self.config.SIZE_A)
  self.MAXARG_Bx = self.MASK_Bx - 1
  self.MAXARG_sBx = math.floor(self.MAXARG_Bx / 2)
  -- field order for encoding
  self.SEQ_iABC = {
    self.config.SIZE_OP, self.config.SIZE_C,
    self.config.SIZE_B,  self.config.SIZE_A,
  }
  ---------------------------------------------------------------
  -- set some implementation constants not defined in global header
  ---------------------------------------------------------------
  -- default constants are defined in llimits.h
  self.config.MAXSTACK = self.header.MAXSTACK or 250
  self.config.MAXVARS = self.header.MAXVARS or 200
  self.config.MAXUPVALUES = self.header.MAXUPVALUES or 32
  self.config.MAXPARAMS = self.header.MAXPARAMS or 100
  self.config.FPF = self.header.FPF or 32
  -- check validity
  if self.config.MAXSTACK < 20 then
    self:Error("MAXSTACK must be greater than 20")
  end
  if self.config.MAXSTACK >= self.MASK_A then
    self:Error("MAXSTACK beyond range of instruction field A")
  end
  if self.config.MAXUPVALUES > self.MASK_Bx then
    self:Error("MAXUPVALUES beyond range of instruction field Bx")
  end
  if self.config.MAXVARS >= self.config.MAXSTACK then
    self:Error("MAXVARS cannot exceed or equal MAXSTACK")
  end
  if self.config.MAXPARAMS >= self.config.MAXVARS then
    self:Error("MAXPARAMS cannot exceed or equal MAXVARS")
  end
  self.config.LUA_TNIL    = 0   -- these can't be touched for now
  self.config.LUA_TNUMBER = 3
  self.config.LUA_TSTRING = 4
end

-----------------------------------------------------------------------
-- Process setting of function properties
-- * nups and list sizes are not handled, they are auto-generated
-----------------------------------------------------------------------
function Code:FunctionOption(pkey, pval)
  self.state = 1
  pkey = string.lower(pkey)
  ---------------------------------------------------------------
  -- convenient test functions for .function property declarations
  ---------------------------------------------------------------
  local function Expect(typ)
    if type(pval) ~= typ then
      self:Error(typ.." expected for property "..pkey)
    end
    if typ == "number" and (math.floor(pval) ~= pval or pval < 0) then
      self:Error("illegal number value for property "..pkey)
    end
  end
  ---------------------------------------------------------------
  local function Range(lo, hi)
    local n = pval
    if math.floor(n) ~= n then
      self:Error("integer number expected for property "..pkey)
    elseif (lo and n < lo) or (hi and n > hi) then
      self:Error("integer out of range for property "..pkey)
    end
  end
  ---------------------------------------------------------------
  -- match, check and set properties
  ---------------------------------------------------------------
  if pkey == "source_name" then
    Expect("string")
    self.funcop.source_name = pval
  ---------------------------------------------------------------
  elseif pkey == "line_defined" then
    Expect("number"); Range(1)
    self.funcop.line_defined = pval
  ---------------------------------------------------------------
  elseif pkey == "numparams" then
    Expect("number"); Range(0, self.config.MAXPARAMS)
    self.funcop.numparams = pval
  ---------------------------------------------------------------
  elseif pkey == "is_vararg" then
    Expect("number"); Range(0, 1)
    self.funcop.is_vararg = pval
  ---------------------------------------------------------------
  elseif pkey == "maxstacksize" then
    Expect("number"); Range(2, self.config.MAXSTACK)
    self.funcop.maxstacksize = pval
  ---------------------------------------------------------------
  else
    self:Error("unrecognized .function property "..pkey)
  end--if pkey
end

-----------------------------------------------------------------------
-- Encode fields into an instruction
-- * OP can be a string
-----------------------------------------------------------------------
function Code:iABC(OP, A, B, C)
  if type(OP) == "string" then OP = Lex.OPCODE[OP] end
  local field = {OP, C, B, A}
  local v, i = "", 0
  local cValue, cBits, cPos = 0, 0, 1
  -- encode an instruction
  while i < self.config.Instruction do
    -- if need more bits, suck in a field at a time
    while cBits < 8 do
      cValue = field[cPos] * math.ldexp(1, cBits) + cValue
      cBits = cBits + self.SEQ_iABC[cPos]; cPos = cPos + 1
    end
    -- extract bytes to instruction string
    while cBits >= 8 do
      v = v..string.char(math.mod(cValue, 256))
      cValue = math.floor(cValue / 256)
      cBits = cBits - 8; i = i + 1
    end
  end
  return v
end

-- not the most efficient methods...
function Code:iABx(OP, A, Bx)
  return self:iABC(OP, A, math.floor(Bx / self.MASK_C),
                          math.mod(Bx, self.MASK_C))
end
function Code:iAsBx(OP, A, sBx)
  -- if forward reference, defer instruction generation
  if type(sBx) == "string" then return {OP, A, sBx} end
  local Bx = sBx + self.MAXARG_sBx
  return self:iABC(OP, A, math.floor(Bx / self.MASK_C),
                          math.mod(Bx, self.MASK_C))
end

-----------------------------------------------------------------------
-- Range checking for operands (other checks in Parse:Instruction)
-----------------------------------------------------------------------
function Code:CheckR(reg)
  if reg >= self.MASK_A then
    self:Error("R-type register operand out of range")
  end
end
function Code:CheckRK(reg)
  if reg >= self.MASK_B then
    self:Error("RK-type register operand out of range")
  end
end
function Code:CheckKst(kst)
  if type(kst) ~= "number" then return end
  if kst >= self.MASK_Bx then
    self:Error("Kst-type constant operand out of range")
  end
end
function Code:CheckDisp(disp)
  if type(disp) ~= "number" then return end
  if disp < -self.MAXARG_sBx or disp > self.MAXARG_sBx then
    self:Error("Displacement-type operand out of range")
  end
end

-----------------------------------------------------------------------
-- Initialize function object
-----------------------------------------------------------------------
function Code:FunctionBegin()
  if self.state == 2 then
    self:Error("cannot have more than one top-level function")
  end
  local func = {}
  ---------------------------------------------------------------
  -- initialize function state
  ---------------------------------------------------------------
  func.locvars = {}             -- init locals table
  func.rlocvars = {}
  func.sizelocvars = 0
  func.localflag = false
  func.upvalues = {}            -- init upvalues table
  func.rupvalues = {}
  func.sizeupvalues = 0
  func.k = {}                   -- init constants table
  func.rk = {}
  func.rksym = {}
  func.rknil = nil
  func.sizek = 0
  func.p = {}                   -- init prototypes table
  func.sizep = 0
  func.code = {}                -- init instructions table
  func.sizecode = 0
  func.lineinfo = {}            -- init debug line info table
  func.maxstacksize = 2
  func.line_defined = Parse.line
  ---------------------------------------------------------------
  -- push one level
  ---------------------------------------------------------------
  self.level = self.level + 1   -- PUSH
  func.prev = self.func         -- save parent struct
  func.prevop = self.funcop     -- save parent function options
  func.prevlb = self.labels     -- save parent label set (2 tables)
  func.prevlt = self.labelt
  self.func = func              -- set this to current
  self.labels = {}              -- new label set
  self.labelt = {}
  self.funcop = {}              -- new function option set
  -- set root (top-level) function if first time
  if not self.root then self.root = func end
end

-----------------------------------------------------------------------
-- Ends and processes function object
-----------------------------------------------------------------------
function Code:FunctionEnd()
  if not self.func then
    self:Error(".end directive without matching .function")
  end
  local parent, now = self.func.prev, self.func
  ---------------------------------------------------------------
  -- resolve forward references for jump displacement labels
  ---------------------------------------------------------------
  if now.sizecode == 0 then
    self:Error("a function cannot have no code")
  end
  for i = 1, now.sizecode do
    local v = now.code[i]
    if type(v) ~= "string" then
      local sym = v[3]
      sym = self:GetLabel(sym, "code")
      if type(sym) ~= "number" then
        self:Error("could not resolve forward reference '"..sym.."'")
      end
      -- displacement relative to the next instruction
      sym = sym - i - 1
      self:CheckDisp(sym)
      now.code[i] = Code:iAsBx(v[1], v[2], sym)
    end
  end
  now.sizelineinfo = now.sizecode       -- these two are the same
  ---------------------------------------------------------------
  -- update sizes of lists (declarations with explicit indices do
  -- not update the size value, so we make sure...)
  -- * TODO there might be holes, not sure what to do with them yet
  ---------------------------------------------------------------
  -- update self.func.sizelocvars
  local maxidx = -1
  for i, v in pairs(now.locvars) do
    if i > maxidx then maxidx = i end
  end
  if maxidx >= now.sizelocvars then now.sizelocvars = maxidx + 1 end
  -- update self.func.sizeupvalues
  maxidx = -1
  for i, v in pairs(now.upvalues) do
    if i > maxidx then maxidx = i end
  end
  if maxidx >= now.sizeupvalues then now.sizeupvalues = maxidx + 1 end
  -- update self.func.sizek
  maxidx = -1
  for i, v in pairs(now.k) do
    if i > maxidx then maxidx = i end
  end
  if maxidx >= now.sizek then now.sizek = maxidx + 1 end
  if now.rknil and now.rknil > maxidx then now.sizek = now.rknil + 1 end
  ---------------------------------------------------------------
  -- set up function header information based on supplied info
  ---------------------------------------------------------------
  now.source_name = self.funcop.source_name
  now.line_defined = self.funcop.line_defined or now.line_defined
  now.nups = now.sizeupvalues
  now.is_vararg = self.funcop.is_vararg or 0
  -- update and validate numparams
  if now.numparams and self.funcop.numparams then
    if self.funcop.numparams < now.numparams then
      self:Error("numparams in .function less than number of .param declarations")
    end
  end
  now.numparams = self.funcop.numparams or now.numparams or 0
  -- update and validate maxstacksize
  if self.funcop.maxstacksize then
    if self.funcop.maxstacksize < now.maxstacksize then
      self:Error("maxstacksize in .function less than required value")
    end
    now.maxstacksize = self.funcop.maxstacksize
  end
  ---------------------------------------------------------------
  -- pop one level
  ---------------------------------------------------------------
  self.level = self.level - 1   -- POP
  self.func = parent            -- load parent struct
  self.funcop = now.prevop      -- load parent function options
  self.labels = now.prevlb      -- load parent label set (2 tables)
  self.labelt = now.prevlt
  if parent then                -- update prototypes table, or
    parent.p[parent.sizep] = now
    parent.sizep = parent.sizep + 1
  else
    self.state = 2              -- top-level is closed properly
    if not now.source_name then -- set the top-level source name
      now.source_name = "@"..self.asmfile
    end
  end
end

-----------------------------------------------------------------------
-- Add instruction to instructions list
-----------------------------------------------------------------------
function Code:AddInst(inst)
  self.func.sizecode = self.func.sizecode + 1
  self.func.code[self.func.sizecode] = inst
  self.func.lineinfo[self.func.sizecode] = self.mneline
end

-----------------------------------------------------------------------
-- Adds a pending label to be noted
-----------------------------------------------------------------------
function Code:AddLabel(tok)
  table.insert(self.labelsnow, tok)
end

-----------------------------------------------------------------------
-- True if there is a label to be processed
-----------------------------------------------------------------------
function Code:HasLabels() return table.getn(self.labelsnow) > 0 end

-----------------------------------------------------------------------
-- Sets a label with a type and a value
-- * if parent flag set, label is added in parent function
--   (this is only for function labels, and self.prev must exist)
-----------------------------------------------------------------------
function Code:SetLabel(typ, val, parent)
  local labelt, labels
  if parent then
    labelt = self.func.prevlt
    labels = self.func.prevlb
  else
    labelt = self.labelt
    labels = self.labels
  end
  for i = 1, table.getn(self.labelsnow) do
    local label = table.remove(self.labelsnow)
    if labelt[label] then
      self:Error("label '"..label.."' cannot be redefined")
    else
      labelt[label] = typ
      labels[label] = val
    end
  end
end

-----------------------------------------------------------------------
-- Gets a given label and type, or return the same symbol if fail
-----------------------------------------------------------------------
function Code:GetLabel(sym, typ)
  idx = self.labels[sym]        -- labels list lookup
  if idx then
    if self.labelt[sym] == typ then
      return idx
    else
      self:Error("'"..sym.."' is not a "..typ.." label")
    end
  end
  -- if not found, it's a forward ref, the parent to defer code gen
  if typ ~= "code" then
    self:Error("label '"..sym.."' cannot be forward-referenced")
  end
  return sym
end

-----------------------------------------------------------------------
-- Gets a function prototype number from a label reference
-----------------------------------------------------------------------
function Code:GetFunction(sym)
  local result = self:GetLabel(sym, "function")
  if result == sym then
    self:Error("function label not found")
  end
  return result
end

-----------------------------------------------------------------------
-- True if symbol is a label
-----------------------------------------------------------------------
function Code:IsLabel(sym) return self.labelt[sym] end

-----------------------------------------------------------------------
-- Sets a local variable
-----------------------------------------------------------------------
function Code:SetLocal(sym, idx)
  self.func.locvars[idx] = sym
  self.func.rlocvars[sym] = idx
  return idx
end

-----------------------------------------------------------------------
-- Adds a local variable (explicit or next location)
-----------------------------------------------------------------------
function Code:AddLocal(sym, idx)
  if not self.func.localflag then
    self.func.localflag = true  -- can no longer accept more params
    self.func.numparams = self.func.sizelocvars
  end
  if idx then -- explicit declaration
    if idx >= self.config.MAXVARS then
      self:Error("local index out of range")
    elseif self.func.locvars[idx] then
      self:Error("local with index "..idx.." already defined")
    end
    return self:SetLocal(sym, idx)
  end
  local i = self.func.sizelocvars -- add in next location
  while i < self.config.MAXVARS do
    if self.func.locvars[i] == nil then
      self.func.sizelocvars = i + 1
      return self:SetLocal(sym, i)
    end
    i = i + 1
  end
  self:Error("too many locals, list overflow")
end

-----------------------------------------------------------------------
-- Adds a parameter (next location), never after a local declaration
-----------------------------------------------------------------------
function Code:AddParam(sym)
  if self.func.localflag then
    self:Error("all parameters must be defined before locals")
  end
  local i = self.func.sizelocvars -- add in the next location
  while i < self.config.MAXPARAMS do
    if self.func.locvars[i] == nil then
      self.func.sizelocvars = i + 1
      self.func.numparams = self.func.sizelocvars
      return self:SetLocal(sym, i)
    end
    i = i + 1
  end
  self:Error("too many parameters, list overflow")
end

-----------------------------------------------------------------------
-- Gets a local or adds a new local
-----------------------------------------------------------------------
function Code:GetLocal(sym)
  local idx = self.func.rlocvars[sym]
  if idx then return idx end
  self:Error("locals must be explicitly declared")
  -- locals should not be added automatically, will cause confusion
  --return self:AddLocal(sym)
end

-----------------------------------------------------------------------
-- Sets an upvalue
-----------------------------------------------------------------------
function Code:SetUpvalue(sym, idx)
  self.func.upvalues[idx] = sym
  self.func.rupvalues[sym] = idx
  return idx
end

-----------------------------------------------------------------------
-- Adds an upvalue (explicit or next location)
-----------------------------------------------------------------------
function Code:AddUpvalue(sym, idx)
  if idx then -- explicit declaration
    if idx >= self.config.MAXUPVALUES then
      self:Error("upvalue index out of range")
    elseif self.func.upvalues[idx] then
      self:Error("upvalue with index "..idx.." already defined")
    end
    return self:SetUpvalue(sym, idx)
  end
  local i = self.func.sizeupvalues -- add in next location
  while i < self.config.MAXUPVALUES do
    if self.func.upvalues[i] == nil then
      self.func.sizeupvalues = i + 1
      return self:SetUpvalue(sym, i)
    end
    i = i + 1
  end
  self:Error("too many upvalues, list overflow")
end

-----------------------------------------------------------------------
-- Checks for an upvalue by its number
-----------------------------------------------------------------------
function Code:IsUpvalue(idx) return self.func.upvalues[idx] end

-----------------------------------------------------------------------
-- Gets an upvalue or adds a new upvalue
-----------------------------------------------------------------------
function Code:GetUpvalue(sym)
  local idx = self.func.rupvalues[sym]
  if idx then return idx end
  return self:AddUpvalue(sym)
end

-----------------------------------------------------------------------
-- Sets a constant
-- * a nil is specially handled since it can't be used as a key
-----------------------------------------------------------------------
function Code:SetConst(sym, val, idx)
  self.func.k[idx] = val
  if val then
    self.func.rk[val] = idx
  else
    self.func.rknil = idx
  end
  if sym then self.func.rksym[sym] = idx end
  return idx
end

-----------------------------------------------------------------------
-- Adds a constant (explicit or next location)
-----------------------------------------------------------------------
function Code:AddConst(sym, val, idx)
  if sym and self.func.rksym[sym] then
    self:Error("constant with symbol '"..sym.."' already defined")
  end
  if idx then -- explicit declaration
    if idx >= self.MAXARG_Bx then
      self:Error("constant index out of range")
    elseif self.func.k[idx] then
      self:Error("constant with index "..idx.." already defined")
    end
    return self:SetConst(sym, val, idx)
  end
  local i = self.func.sizek -- add in next location
  while i < self.MAXARG_Bx do
    if self.func.k[i] == nil and self.func.rknil ~= i then
      self.func.sizek = i + 1
      return self:SetConst(sym, val, i)
    end
    i = i + 1
  end
  self:Error("too many constants, list overflow")
end

-----------------------------------------------------------------------
-- Gets a constant or adds a new constant by symbol
-----------------------------------------------------------------------
function Code:GetConstSym(sym)
  local idx = self.func.rksym[sym]
  if idx then return idx end
end

-----------------------------------------------------------------------
-- Checks for a constant by constant number
-----------------------------------------------------------------------
function Code:IsConst(idx)
  if self.func.rknil == idx then return true end
  return self.func.k[idx]
end

-----------------------------------------------------------------------
-- Gets a constant or adds a new constant by value
-----------------------------------------------------------------------
function Code:GetConst(val)
  if val == nil then
    if self.func.rknil then return self.func.rknil end
    return self:AddConst(nil, nil)
  end
  local idx = self.func.rk[val]
  if idx then return idx end
  return self:AddConst(nil, val)
end

-----------------------------------------------------------------------
-- Dumps a binary chunk of the assembled code as a string
-----------------------------------------------------------------------
function Code:Dump()
  local function Byte(c) return string.char(c) end
  ---------------------------------------------------------------
  -- return a block of bytes with correct endianness
  ---------------------------------------------------------------
  local function Block(v)
    if self.config.endianness == 0 then -- big endian
      local w = ""
      for i = string.len(v), 1, -1 do
        w = w..string.char(string.byte(v, i))
      end
      return w
    else
      return v
    end
  end
  ---------------------------------------------------------------
  -- builds global header
  ---------------------------------------------------------------
  if not self.config or not self.root then
    self:Error("no valid function defined in source file")
  end
  if self.state ~= 2 then
    self:Error("top-level function not yet closed by a .end")
  end
  local header =
    self.config.signature..
    Byte(self.config.version)..
    Byte(self.config.endianness)..
    Byte(self.config.int)..             -- byte sizes
    Byte(self.config.size_t)..
    Byte(self.config.Instruction)..
    Byte(self.config.SIZE_OP)..         -- bit sizes
    Byte(self.config.SIZE_A)..
    Byte(self.config.SIZE_B)..
    Byte(self.config.SIZE_C)..
    Byte(self.config.lua_Number)..      -- number info
    Block(self.config.test_number)
  ---------------------------------------------------------------
  -- recursively called to dump function data
  ---------------------------------------------------------------
  local function DumpFunction(func)
    -------------------------------------------------------------
    -- dumps an unsigned integer (for integers, size_ts)
    -------------------------------------------------------------
    local function DumpUnsigned(num, type_size)
      if not type_size then type_size = self.config.int end
      local v = ""
      for i = 1, type_size do
        v = v..string.char(math.mod(num, 256))
        num = math.floor(num / 256)
      end
      return Block(v)
    end
    -------------------------------------------------------------
    -- dumps a number (lua_Number type)
    -------------------------------------------------------------
    local function DumpNumber(num)
      if not Number.Convert then
        self:Error("conversion function for lua_Number not initialized")
      end
      return Block(Number:Convert(num))
    end
    -------------------------------------------------------------
    -- dumps a string
    -------------------------------------------------------------
    local function DumpString(str)
      if not str then
        return DumpUnsigned(0, config.size_size_t)
      end
      str = str.."\0"   -- mandatory NUL termination
      return DumpUnsigned(string.len(str), self.config.size_t)..str
    end
    -------------------------------------------------------------
    -- dumps debug line information
    -------------------------------------------------------------
    local function DumpLines()
      local s = DumpUnsigned(func.sizelineinfo)
      if func.sizelineinfo == 0 then return s end
      --TODO debug info option
      for i = 1, func.sizelineinfo do
        s = s..DumpUnsigned(func.lineinfo[i])
      end
      return s
    end
    -------------------------------------------------------------
    -- dump local variables
    -------------------------------------------------------------
    local function DumpLocals()
      local s = DumpUnsigned(func.sizelocvars)
      if func.sizelocvars == 0 then return s end
      for i = 0, func.sizelocvars - 1 do
        local locname = func.locvars[i]
        if not locname then locname = "(none)" end
        local startpc, endpc = 1, func.sizecode
        if func.numparams > i then startpc = 0 end
        s = s..DumpString(locname)..
               DumpUnsigned(startpc)..
               DumpUnsigned(endpc)
      end
      return s
    end
    -------------------------------------------------------------
    -- dump upvalues
    -------------------------------------------------------------
    local function DumpUpvalues()
      local s = DumpUnsigned(func.sizeupvalues)
      if func.sizeupvalues == 0 then return s end
      for i = 0, func.sizeupvalues - 1 do
        local upval = func.upvalues[i]
        if not upval then upval = "(none)" end
        s = s..DumpString(upval)
      end
      return s
    end
    -------------------------------------------------------------
    -- dump constants
    -- * generates nil constants; Lua doesn't seem to do that...
    -------------------------------------------------------------
    local function DumpConstantKs()
      local s = DumpUnsigned(func.sizek)
      if func.sizek == 0 then return s end
      for i = 0, func.sizek -1 do
        local v = func.k[i]
        if type(v) == "number" then
          s = s..Byte(self.config.LUA_TNUMBER)..DumpNumber(v)
        elseif type(v) == "string" then
          s = s..Byte(self.config.LUA_TSTRING)..DumpString(v)
        elseif type(v) == "nil" then
          s = s..Byte(self.config.LUA_TNIL)
        else
          self:Error("failed writing constant "..i.." bad type")
        end
      end--for
      return s
    end
    -------------------------------------------------------------
    -- dump function prototypes
    -------------------------------------------------------------
    local function DumpConstantPs()
      local s = DumpUnsigned(func.sizep)
      if func.sizep == 0 then return s end
      for i = 0, func.sizep - 1 do
        s = s..DumpFunction(func.p[i])
      end
      return s
    end
    -------------------------------------------------------------
    -- dump code
    -------------------------------------------------------------
    local function DumpCode()
      local s = DumpUnsigned(func.sizecode)
      if func.sizecode == 0 then return s end
      for i = 1, func.sizecode do
        s = s..Block(func.code[i])
      end
      return s
    end
    -------------------------------------------------------------
    -- body of WriteFunction() starts here
    -------------------------------------------------------------
    return DumpString(func.source_name)..
           DumpUnsigned(func.line_defined)..
           Byte(func.nups)..            -- some byte counts
           Byte(func.numparams)..
           Byte(func.is_vararg)..
           Byte(func.maxstacksize)..
           DumpLines()..                -- these are lists
           DumpLocals()..
           DumpUpvalues()..
           DumpConstantKs()..
           DumpConstantPs()..           -- may be recursive
           DumpCode()
  end--DumpFunction

  ---------------------------------------------------------------
  -- recursive call to function writing process
  ---------------------------------------------------------------
  return header..DumpFunction(self.root)
end

--[[-------------------------------------------------------------------
-- Top-level file handling, processes user's input file list
--]]-------------------------------------------------------------------

function ChunkBakeDoFiles(files)
  for i, asmfile in ipairs(files) do
    local binfile, lstfile
    -------------------------------------------------------------
    -- find and replace extension for filenames
    -------------------------------------------------------------
    local extb, exte = string.find(asmfile, "%.[^%.%\\%/]*$")
    local basename = asmfile
    if extb and extb > 1 then
      basename = string.sub(asmfile, 1, extb - 1)
    end
    if config.LISTING_FLAG then lstfile = basename..config.EXT_LST end
    binfile = config.OUTPUT_FILE or basename..config.EXT_BIN
    -------------------------------------------------------------
    -- load and parse (source) assembly files
    -------------------------------------------------------------
    local INF = io.open(asmfile, "rb")
    if not INF then
      error("cannot open \""..asmfile.."\" for reading")
    end
    local asmdata = INF:read("*a")
    if not asmdata then
      error("error reading \""..asmfile.."\" assembly file")
    end
    io.close(INF)
    -- assembly starts from here...
    local bindata, lstdata = Parse:Parse(asmdata, asmfile)
    -------------------------------------------------------------
    -- optionally write out binary file
    -------------------------------------------------------------
    if not config.RUN_FLAG and binfile then
      local OUTF = io.open(binfile, "wb")
      if not OUTF then
        error("cannot open \""..binfile.."\" for writing")
      end
      OUTF:write(bindata)
      io.close(OUTF)
    end
--[[
    -------------------------------------------------------------
    --TODO OPTIONALLY write out listing file
    -------------------------------------------------------------
    if not config.RUN_FLAG and lstfile then
    end
    -------------------------------------------------------------
    --TODO OPTIONALLY run the binary chunk
    -------------------------------------------------------------
    if config.RUN_FLAG then
    end
--]]
  end--for
end

--[[-------------------------------------------------------------------
-- Command-line interface
--]]-------------------------------------------------------------------

function main()
  ---------------------------------------------------------------
  -- handle arguments
  ---------------------------------------------------------------
  if table.getn(arg) == 0 then
    print(title) print(usage) return
  end
  local files, i = {}, 1
  while i <= table.getn(arg) do
    local a, b = arg[i], arg[i + 1]
    if string.sub(a, 1, 1) == "-" then        -- handle options here
      if a == "-h" or a == "--help" then
        print(title) print(usage) return
      elseif a == "--quiet" then
        config.QUIET = true
--[[
      elseif a == "--list" then
        config.LISTING_FLAG = true
--]]
      elseif a == "-o" then
        if not b then error("-o option needs a file name") end
        config.OUTPUT_FILE = b
        i = i + 1
--[[
      elseif a == "--run" then
        config.RUN_FLAG = true
--]]
      elseif a == "--" then
        for j = i + 1, table.getn(arg) do
          table.insert(arg_other, arg[j])     -- gather rest of args
        end
        break
      else
        error("unrecognized option "..a)
      end
    else
      table.insert(files, a)                  -- potential filename
    end
    i = i + 1
  end--while
  if table.getn(files) > 0 then
    if table.getn(files) > 1 then
      if config.OUTPUT_FILE then
        error("with -o, only one source file can be specified")
      elseif config.RUN_FLAG then
        error("specify only one assembly file with --run")
      end
    end
    ChunkBakeDoFiles(files)
  else
    print(title) print("ChunkBake: nothing to do!")
  end
end

-----------------------------------------------------------------------
-- program entry point
-----------------------------------------------------------------------
if not TEST then
  local ok, msg = pcall(main)           -- call main() for proper traceback
  if not ok then                        -- error
    print(title)
    print("* Run with option -h or --help for usage information")
    print(msg)
  end
end

----end-of-script----
